"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.QueryProxyMock = void 0;
const axios_1 = __importDefault(require("axios"));
const buffer_1 = require("buffer");
const query_1 = require("../query");
const BinaryWriter_1 = require("../query/BinaryWriter");
const keccak256_1 = require("@ethersproject/keccak256");
const QUERY_RESPONSE_PREFIX = "query_response_0000000000000000000|";
/**
 * Usage:
 *
 * ```js
 * const mock = new QueryProxyMock({
 *   2: "http://localhost:8545",
 * });
 * ```
 *
 * If you are running an Anvil fork like
 *
 * ```bash
 * anvil -f https://ethereum-goerli.publicnode.com
 * ```
 *
 * You can use the following command to switch the guardian address to the devnet / mock guardian
 *
 * Where the `-a` parameter is the core bridge address on that chain
 *
 * https://docs.wormhole.com/wormhole/reference/constants#core-contracts
 *
 * ```bash
 * npx \@wormhole-foundation/wormhole-cli evm hijack -a 0x706abc4E45D419950511e474C7B9Ed348A4a716c -g 0xbeFA429d57cD18b7F8A4d91A2da9AB4AF05d0FBe
 * ```
 */
class QueryProxyMock {
    constructor(rpcMap, mockPrivateKeys = [
        "cfb12303a19cde580bb4dd771639b0d26bc68353645571a8cff516ab2ee113a0",
    ]) {
        this.rpcMap = rpcMap;
        this.mockPrivateKeys = mockPrivateKeys;
    }
    sign(serializedResponse) {
        const digest = (0, query_1.hexToUint8Array)((0, keccak256_1.keccak256)(buffer_1.Buffer.concat([
            buffer_1.Buffer.from(QUERY_RESPONSE_PREFIX),
            (0, query_1.hexToUint8Array)((0, keccak256_1.keccak256)(serializedResponse)),
        ])));
        return this.mockPrivateKeys.map((key, idx) => `${(0, query_1.sign)(key, digest)}${idx.toString(16).padStart(2, "0")}`);
    }
    /**
     * Usage:
     *
     * ```js
     * const { bytes, signatures } = await mock.mock(new QueryRequest(nonce, [
     *   new PerChainQueryRequest(
     *     wormholeChainId,
     *     new EthCallQueryRequest(blockNumber, [
     *       { to: contract, data: abiEncodedData },
     *     ])
     *   ),
     * ]));
     * ```
     *
     * @param queryRequest an instance of the `QueryRequest` class
     * @returns a promise result matching the query proxy's query response
     */
    mock(queryRequest) {
        var _a, _b, _c, _d, _e, _f, _g, _h;
        return __awaiter(this, void 0, void 0, function* () {
            const serializedRequest = queryRequest.serialize();
            const writer = new BinaryWriter_1.BinaryWriter()
                .writeUint8(1) // version
                .writeUint16(0) // source = off-chain
                .writeUint8Array(new Uint8Array(new Array(65))) // empty signature for mock
                .writeUint32(serializedRequest.length)
                .writeUint8Array(serializedRequest)
                .writeUint8(queryRequest.requests.length);
            for (const perChainRequest of queryRequest.requests) {
                const rpc = this.rpcMap[perChainRequest.chainId];
                if (!rpc) {
                    throw new Error(`Unregistered chain id for mock: ${perChainRequest.chainId}`);
                }
                const type = perChainRequest.query.type();
                writer.writeUint16(perChainRequest.chainId).writeUint8(type);
                if (type === query_1.ChainQueryType.EthCall) {
                    const query = perChainRequest.query;
                    const response = yield axios_1.default.post(rpc, [
                        ...query.callData.map((args, idx) => ({
                            jsonrpc: "2.0",
                            id: idx,
                            method: "eth_call",
                            params: [
                                args,
                                //TODO: support block hash
                                query.blockTag,
                            ],
                        })),
                        {
                            jsonrpc: "2.0",
                            id: query.callData.length,
                            //TODO: support block hash
                            method: "eth_getBlockByNumber",
                            params: [query.blockTag, false],
                        },
                    ]);
                    const callResults = (_a = response === null || response === void 0 ? void 0 : response.data) === null || _a === void 0 ? void 0 : _a.slice(0, query.callData.length);
                    const blockResult = (_c = (_b = response === null || response === void 0 ? void 0 : response.data) === null || _b === void 0 ? void 0 : _b[query.callData.length]) === null || _c === void 0 ? void 0 : _c.result;
                    if (!blockResult ||
                        !blockResult.number ||
                        !blockResult.timestamp ||
                        !blockResult.hash) {
                        throw new Error(`Invalid block result for chain ${perChainRequest.chainId} block ${query.blockTag}`);
                    }
                    const results = callResults.map((callResult) => new Uint8Array(buffer_1.Buffer.from(callResult.result.substring(2), "hex")));
                    const perChainWriter = new BinaryWriter_1.BinaryWriter()
                        .writeUint64(BigInt(parseInt(blockResult.number.substring(2), 16))) // block number
                        .writeUint8Array(new Uint8Array(buffer_1.Buffer.from(blockResult.hash.substring(2), "hex"))) // hash
                        .writeUint64(BigInt(parseInt(blockResult.timestamp.substring(2), 16)) *
                        BigInt("1000000")) // time in seconds -> microseconds
                        .writeUint8(results.length);
                    for (const result of results) {
                        perChainWriter.writeUint32(result.length).writeUint8Array(result);
                    }
                    const serialized = perChainWriter.data();
                    writer.writeUint32(serialized.length).writeUint8Array(serialized);
                }
                else if (type === query_1.ChainQueryType.EthCallWithFinality) {
                    const query = perChainRequest.query;
                    const response = yield axios_1.default.post(rpc, [
                        ...query.callData.map((args, idx) => ({
                            jsonrpc: "2.0",
                            id: idx,
                            method: "eth_call",
                            params: [
                                args,
                                //TODO: support block hash
                                query.blockId,
                            ],
                        })),
                        {
                            jsonrpc: "2.0",
                            id: query.callData.length,
                            //TODO: support block hash
                            method: "eth_getBlockByNumber",
                            params: [query.blockId, false],
                        },
                        {
                            jsonrpc: "2.0",
                            id: query.callData.length,
                            method: "eth_getBlockByNumber",
                            params: [query.finality, false],
                        },
                    ]);
                    const callResults = (_d = response === null || response === void 0 ? void 0 : response.data) === null || _d === void 0 ? void 0 : _d.slice(0, query.callData.length);
                    const blockResult = (_f = (_e = response === null || response === void 0 ? void 0 : response.data) === null || _e === void 0 ? void 0 : _e[query.callData.length]) === null || _f === void 0 ? void 0 : _f.result;
                    const finalityBlockResult = (_h = (_g = response === null || response === void 0 ? void 0 : response.data) === null || _g === void 0 ? void 0 : _g[query.callData.length + 1]) === null || _h === void 0 ? void 0 : _h.result;
                    if (!blockResult ||
                        !blockResult.number ||
                        !blockResult.timestamp ||
                        !blockResult.hash) {
                        throw new Error(`Invalid block result for chain ${perChainRequest.chainId} block ${query.blockId}`);
                    }
                    if (!finalityBlockResult ||
                        !finalityBlockResult.number ||
                        !finalityBlockResult.timestamp ||
                        !finalityBlockResult.hash) {
                        throw new Error(`Invalid tagged block result for chain ${perChainRequest.chainId} tag ${query.finality}`);
                    }
                    const blockNumber = BigInt(parseInt(blockResult.number.substring(2), 16));
                    const latestBlockNumberMatchingFinality = BigInt(parseInt(finalityBlockResult.number.substring(2), 16));
                    if (blockNumber > latestBlockNumberMatchingFinality) {
                        throw new Error(`Requested block for eth_call_with_finality has not yet reached the requested finality. Block: ${blockNumber}, ${query.finality}: ${latestBlockNumberMatchingFinality}`);
                    }
                    const results = callResults.map((callResult) => new Uint8Array(buffer_1.Buffer.from(callResult.result.substring(2), "hex")));
                    const perChainWriter = new BinaryWriter_1.BinaryWriter()
                        .writeUint64(blockNumber) // block number
                        .writeUint8Array(new Uint8Array(buffer_1.Buffer.from(blockResult.hash.substring(2), "hex"))) // hash
                        .writeUint64(BigInt(parseInt(blockResult.timestamp.substring(2), 16)) *
                        BigInt("1000000")) // time in seconds -> microseconds
                        .writeUint8(results.length);
                    for (const result of results) {
                        perChainWriter.writeUint32(result.length).writeUint8Array(result);
                    }
                    const serialized = perChainWriter.data();
                    writer.writeUint32(serialized.length).writeUint8Array(serialized);
                }
                else {
                    throw new Error(`Unsupported query type for mock: ${type}`);
                }
            }
            const serializedResponse = writer.data();
            return {
                signatures: this.sign(serializedResponse),
                bytes: buffer_1.Buffer.from(serializedResponse).toString("hex"),
            };
        });
    }
}
exports.QueryProxyMock = QueryProxyMock;
